import subprocess
import tempfile
from rich.live import Live
from typing import List, Optional


def _run_fzf(choices: List[str], live: Optional[Live]) -> List[str]:
    choices.reverse()  # reverse order when searching - from bottom to top
    choices_str = "\n".join(choices)
    selection = []
    if live:
        live.stop()
    fzf_args = filter(None, ["fzf", "--ansi"])
    with tempfile.NamedTemporaryFile() as input_file:
        with tempfile.NamedTemporaryFile() as output_file:
            input_file.write(choices_str.encode("utf-8"))
            input_file.flush()
            cat = subprocess.Popen(["cat", input_file.name], stdout=subprocess.PIPE)
            subprocess.run(fzf_args, stdin=cat.stdout, stdout=output_file)
            cat.wait()
            with open(output_file.name) as f:
                selection = [line.strip("\n") for line in f.readlines()]
    if live:
        live.start()
    return selection


def pick_one(choices: List[str], live: Optional[Live]) -> Optional[str]:
    result = next(iter(_run_fzf(choices, live)), None)
    return result



import webbrowser
import wsgiref.simple_server
import wsgiref.util

from google_auth_oauthlib.flow import _WSGIRequestHandler, Flow, _RedirectWSGIApp
import requests
from rich.console import Console
from rich.text import Text

from bq_meta import const
from bq_meta.config import Config


class Auth:
    def __init__(self, config: Config, console: Console) -> None:
        self.config = config
        self.console = console

    def login(self):
        flow = Flow.from_client_config(self.config.client_config, scopes=self.config.scopes)
        try:
            webbrowser.get()
            host = "localhost"
            wsgi_app = _RedirectWSGIApp("The authentication flow has completed. You may close this window.")
            wsgi_app.allow_reuse_address = False
            local_server = wsgiref.simple_server.make_server(host, 0, wsgi_app, handler_class=_WSGIRequestHandler)
            flow.redirect_uri = f"http://{host}:{local_server.server_port}/"
            auth_url, _ = flow.authorization_url()
            webbrowser.open(auth_url, new=2)
            self.console.print(
                Text("Browser openned", style=const.info_style).append(f": {auth_url}", style=const.darker_style)
            )
            local_server.handle_request()
            authorization_response = wsgi_app.last_request_uri.replace("http", "https")
            flow.fetch_token(authorization_response=authorization_response)
            local_server.server_close()
        except Exception:
            flow.redirect_uri = "urn:ietf:wg:oauth:2.0:oob"
            auth_url, _ = flow.authorization_url()
            self.console.print(
                Text("Open following link", style=const.request_style).append(f": {auth_url}", style=const.link_style),
            )
            code = self.console.input(
                Text("Enter verification code", style=const.request_style).append(f": ", style=const.darker_style)
            )
            flow.fetch_token(code=code)

        response = requests.get(
            "https://www.googleapis.com/oauth2/v1/userinfo",
            headers={"Authorization": f"Bearer {flow.credentials.token}"},
        )
        self.config.account = response.json()["email"]
        self.config.credentials = flow.credentials.to_json()
        self.console.print(Text("Credentials updated", style=const.info_style))



import json
from typing import List

import yaml

from bq_meta import const


class Config:
    default = {
        "client_config": {  # client_config of the google-cloud-sdk
            "installed": {
                "client_id": "32555940559.apps.googleusercontent.com",
                "client_secret": "ZmssLNjJy2998hD4CTg2ejr2",
                "auth_uri": "https://accounts.google.com/o/oauth2/auth",
                "token_uri": "https://oauth2.googleapis.com/token",
            }
        },
        "scopes": [
            "https://www.googleapis.com/auth/bigquery",
            "https://www.googleapis.com/auth/userinfo.email",
            "openid",
        ],
        "credentials": None,
        "current_version": "",
        "available_version": "",
        "account": "",
    }

    def __init__(self) -> None:
        self.config_path = const.BQ_META_CONFIG
        self._conf = None

    def write_default(self):
        conf = Config.default
        self._save_conf(conf)

    def _save_conf(self, conf: dict):
        self._conf = conf
        with open(self.config_path, "w") as f:
            yaml.safe_dump(conf, f)

    @property
    def conf(self) -> dict:
        if not self._conf:
            with open(self.config_path, "r") as f:
                self._conf = yaml.safe_load(f)
        return self._conf

    @property
    def client_config(self) -> dict:
        return self.conf["client_config"]

    @property
    def scopes(self) -> List[str]:
        return self.conf["scopes"]

    @property
    def account(self) -> str:
        return self.conf["account"]

    @account.setter
    def account(self, account: str):
        conf = {**self.conf, "account": account}
        self._save_conf(conf)

    @property
    def credentials(self) -> dict:
        return json.loads(self.conf["credentials"])

    @credentials.setter
    def credentials(self, credentials: dict):
        conf = {**self.conf, "credentials": credentials}
        self._save_conf(conf)

    @property
    def current_version(self) -> str:
        return self.conf.get("current_version", None)

    @current_version.setter
    def current_version(self, current_version: str):
        conf = {**self.conf, "current_version": current_version}
        self._save_conf(conf)

    @property
    def available_version(self) -> str:
        return self.conf.get("available_version", None)

    @available_version.setter
    def available_version(self, available_version: str):
        conf = {**self.conf, "available_version": available_version}
        self._save_conf(conf)



from threading import Event
from typing import Iterator
from rich.console import Console

from rich.progress import Progress, TextColumn, TimeElapsedColumn
from rich.live import Live
from rich.panel import Panel
from rich.layout import Layout

from bq_meta import const


def progress(console: Console, name: str, iterator: Iterator) -> list:
    size = 0
    result = []
    with Progress(
        TextColumn("[progress.description]{task.description}", style=const.info_style),
        "•",
        TextColumn("[progress.completed]{task.completed}"),
        "•",
        TimeElapsedColumn(),
        console=console,
        transient=True,
    ) as progress:
        task = progress.add_task(f"Fetching {name}")
        for item in iterator:
            result.append(item)
            progress.advance(task)
            size += 1
    return result


def flash_panel(live: Live, layout: Layout, panel: Panel):
    event = Event()
    panel.border_style = const.request_style  # border will flash a short period of time
    live.update(layout, refresh=True)
    event.wait(0.1)
    if not event.is_set():
        panel.border_style = const.border_style
        live.update(layout, refresh=True)



from google.cloud.bigquery import Client
from google.oauth2.credentials import Credentials
from rich.console import Console
from rich.text import Text

from bq_meta import const
from bq_meta.config import Config


class BqClient:
    def __init__(self, console: Console, config: Config):
        self._client = None
        self.console = console
        self.config = config

    @property
    def client(self) -> Client:
        if not self._client:
            with self.console.status(Text("Connecting to the API", style=const.darker_style), spinner="point"):
                self._client = Client(
                    project="",  # leaving project as None would result in error
                    credentials=Credentials.from_authorized_user_info(self.config.credentials),
                )
        return self._client



import locale
from typing import Optional

locale.setlocale(locale.LC_ALL, "")


def bytes_fmt(bytes_opt: Optional[int]) -> str:
    bytes = bytes_opt if bytes_opt else 0
    for unit in ["B", "KiB", "MiB", "GiB", "TiB", "PiB"]:
        if abs(bytes) < 1024:
            size = round(bytes, 2)
            return f"{size} {unit}"
        else:
            bytes /= 1024


def num_fmt(num_opt: Optional[int]) -> str:
    num = num_opt if num_opt else 0
    return f"{num:,}"



from google.cloud import bigquery
from packaging import version
from rich.align import Align
from rich.box import SIMPLE
from rich.columns import Columns
from rich.console import Group, NewLine
from rich.layout import Layout
from rich.rule import Rule
from rich.table import Table
from rich.text import Text
from rich.tree import Tree

from bq_meta import const
from bq_meta.config import Config
from bq_meta.util import table_utils
from bq_meta.util.num_utils import bytes_fmt, num_fmt

title = """
██▄ █ ▄▀  ▄▀▄ █ █ ██▀ █▀▄ ▀▄▀   █▄ ▄█ ██▀ ▀█▀ ▄▀▄ █▀▄ ▄▀▄ ▀█▀ ▄▀▄
█▄█ █ ▀▄█ ▀▄█ ▀▄█ █▄▄ █▀▄  █    █ ▀ █ █▄▄  █  █▀█ █▄▀ █▀█  █  █▀█
"""


def header_layout(config: Config) -> Layout:
    header_title = Align(title, align="center", style=const.info_style)
    header = Layout(name="header", size=4)
    left = version_layout(config)
    mid = Layout(header_title)
    right = Layout(NewLine(), size=20)
    header.split_row(left, mid, right)
    return header


def version_layout(config: Config) -> Layout:
    if config.current_version and config.available_version:
        current = version.parse(config.current_version)
        available = version.parse(config.available_version)
        if current < available:
            version_text = Text(f"{current} ► {available}", style=const.darker_style)
        else:
            version_text = Text(f"{current}", style=const.darker_style)
    else:
        version_text = Text(" ", style=const.darker_style)
    return Layout(version_text, size=20)


def get_config_info(config: Config) -> Group:
    return Group(text_tuple("Account", config.account))


def get_schema_output(table: bigquery.Table) -> Group:
    schema = table.schema
    tree = Tree(Text("Schema", const.key_style))
    table = Table(box=SIMPLE, show_header=False)
    table_utils.scheme_tree(schema, tree)
    table_utils.scheme_table(schema, table)
    return Group(Rule(style=const.darker_style), Columns([tree, table]))


# fmt: off
def get_table_output(table: bigquery.Table) -> Group:
    total_logical_bytes = bytes_fmt(int(table._properties.get("numTotalLogicalBytes", 0)))
    active_logical_bytes = bytes_fmt(int(table._properties.get("numActiveLogicalBytes", 0)))
    long_term_logical_bytes = bytes_fmt(int(table._properties.get("numLongTermLogicalBytes", 0)))
    total_physical_bytes = bytes_fmt(int(table._properties.get("numTotalPhysicalBytes", 0)))
    active_physical_bytes = bytes_fmt(int(table._properties.get("numActivePhysicalBytes", 0)))
    long_term_physical_bytes = bytes_fmt(int(table._properties.get("numLongTermPhysicalBytes", 0)))
    time_travel_physical_bytes = bytes_fmt(int(table._properties.get("numTimeTravelPhysicalBytes", 0)))
    rows = num_fmt(table.num_rows)
    
    created = table.created.strftime("%Y-%m-%d %H:%M:%S UTC")
    modified = table.modified.strftime("%Y-%m-%d %H:%M:%S UTC")
    expiry = table.expires.strftime("%Y-%m-%d %H:%M:%S UTC") if table.expires else None
    
    partitioned_by = table.time_partitioning.type_ if table.time_partitioning else None
    partitioned_field = table.time_partitioning.field if table.time_partitioning else None
    partition_filter = table.require_partition_filter if table.require_partition_filter else None
    _num_partitions = table._properties.get("numPartitions", None)
    num_of_partitions = int(_num_partitions) if _num_partitions else None
    
    streaming_buffer_size = bytes_fmt(table.streaming_buffer.estimated_bytes) if table.streaming_buffer else None
    streaming_buffer_rows = num_fmt(table.streaming_buffer.estimated_rows) if table.streaming_buffer else None
    streaming_entry_time = table.streaming_buffer.oldest_entry_time.strftime("%Y-%m-%d %H:%M:%S UTC") if table.streaming_buffer else None
    return Group(
        Rule("Table info", style=const.darker_style),
        table_tuple({
            "Table ID": table.full_table_id,
            "Friendly name": table.friendly_name,
            "Created": created,
            "Last modified": modified,
            "Table expiry": expiry,
            "Data location": table.location,
            "Data location": table.location,
        }),
        Rule(style=const.darker_style),
        table_tuple({
            "Table type": table.table_type,
            "Partitioned by": partitioned_by,
            "Partitioned on field": partitioned_field,
            "Partition expiry": table.partition_expiration,
            "Partition filter": partition_filter,
            "Clustered by": table.clustering_fields,
        }),
        Rule("Storage info", style=const.darker_style),
        table_tuple({
            "Number of row": rows, 
            "Number of partitions": num_of_partitions, 
            "Total logical bytes": total_logical_bytes, 
            "Active logical bytes": active_logical_bytes, 
            "Long-term logical bytes": long_term_logical_bytes, 
            "Total physical bytes": total_physical_bytes, 
            "Active physical bytes": active_physical_bytes, 
            "Long-term physical bytes": long_term_physical_bytes, 
            "Time travel physical bytes": time_travel_physical_bytes, 
        }),
        Rule("Streaming buffer statistics", style=const.darker_style),
        table_tuple({
            "Estimated size": streaming_buffer_size,
            "Estimated rows": streaming_buffer_rows,
            "Earliest entry time": streaming_entry_time,
        }),
        Rule(style=const.darker_style),
    )
# fmt: on


def table_tuple(tuples: dict) -> Text:
    table = Table(
        box=const.equal_box,
        show_header=False,
        show_edge=False,
        pad_edge=False,
        border_style=const.darker_style,
    )
    for key, value in tuples.items():
        text = value if isinstance(tuple, Text) else Text(str(value), style="default")
        table.add_row(Text(key, style=const.key_style), text)
    return table


def text_tuple(name: str, value) -> Text:
    text = None
    if isinstance(value, Text):
        text = value
    else:
        text = Text(str(value), style="default")
    return Text(name, style=const.key_style).append(" = ", style=const.darker_style).append(text)



from typing import List

from bq_meta import const
from bq_meta.bq_client import BqClient
from bq_meta.config import Config
from rich.console import Console

from bq_meta.util.rich_utils import progress


class ProjectService:
    def __init__(self, console: Console, config: Config, bq_client: BqClient) -> None:
        self.console = console
        self.config = config
        self.bq_client = bq_client
        self.projects_path = const.BQ_META_PROJECTS

    def list_projects(self) -> List[str]:
        projects = open(self.projects_path, "r").read().splitlines()
        return projects

    def fetch_projects(self):
        projects_ids = ["bigquery-public-data"]  # add bq public datasets explicitly, for testing purposes
        iterator = self.bq_client.client.list_projects()
        for project in progress(self.console, "projects", iterator):
            projects_ids.append(project.project_id)

        with open(self.projects_path, "w") as f:
            f.write("\n".join(projects_ids))



from bq_meta.config import Config
import requests
from importlib.metadata import version


class VersionService:
    def __init__(self):
        ...

    def update_config(self, config: Config):
        available_version = VersionService._get_available_version()
        current_version = VersionService._get_current_version()
        config.available_version = available_version
        config.current_version = current_version

    def _get_available_version() -> str:
        package = "bq-meta"
        url = f"https://pypi.org/pypi/{package}/json"
        response: dict = requests.request("GET", url).json()
        version = response.get("info", None).get("version", None)
        return version

    def _get_current_version() -> str:
        try:
            ver = version("bq-meta")
        except:
            ver = None
        return ver




from typing import Optional

from bq_meta.bq_client import BqClient
from bq_meta.config import Config
from bq_meta.service.project_service import ProjectService
from bq_meta.util import bash_util
from bq_meta.util.rich_utils import progress
from google.cloud import bigquery
from rich.console import Console
from rich.live import Live


class TableService:
    def __init__(
        self,
        console: Console,
        config: Config,
        bq_client: BqClient,
        project_service: ProjectService,
    ):
        self.console = console
        self.config = config
        self.bq_client = bq_client
        self.project_service = project_service

    def get_table(
        self,
        project_id: Optional[str] = None,
        dataset_id: Optional[str] = None,
        table_id: Optional[str] = None,
        live: Optional[Live] = None,
    ) -> Optional[bigquery.Table]:
        table = None
        project_id = project_id if project_id else self._pick_project_id(live)
        if project_id:
            dataset_id = dataset_id if dataset_id else self._pick_dataset_id(project_id, live)
        if dataset_id:
            table_id = table_id if table_id else self._pick_table_id(project_id, dataset_id, live)
        if table_id:
            table = self.bq_client.client.get_table(f"{project_id}.{dataset_id}.{table_id}")
        return table

    def get_fresh_table(self, table: bigquery.Table) -> bigquery.Table:
        return self.bq_client.client.get_table(f"{table.project}.{table.dataset_id}.{table.table_id}")

    # ======================   Pick   ======================

    def _pick_project_id(self, live: Optional[Live]) -> Optional[str]:
        project_ids = self.project_service.list_projects()
        return bash_util.pick_one(project_ids, live)

    def _pick_dataset_id(self, project_id: str, live: Optional[Live]) -> Optional[str]:
        dataset_ids = []
        iterator = self.bq_client.client.list_datasets(project=project_id)
        for dataset in progress(self.console, "datasets", iterator):
            dataset_ids.append(dataset.dataset_id)
        return bash_util.pick_one(dataset_ids, live)

    def _pick_table_id(self, project_id: str, dataset_id: str, live: Optional[Live]) -> Optional[str]:
        table_ids = []
        iterator = self.bq_client.client.list_tables(f"{project_id}.{dataset_id}")
        for table in progress(self.console, "tables", iterator):
            table_ids.append(table.table_id)
        return bash_util.pick_one(table_ids, live)



import os
from typing import Optional

import click
from rich.console import Console
from rich.panel import Panel
from rich.text import Text
from google.cloud.bigquery.table import TableReference

from bq_meta import const, output
from bq_meta.bq_client import BqClient
from bq_meta.config import Config
from bq_meta.initialize import initialize
from bq_meta.service.history_service import HistoryService
from bq_meta.service.project_service import ProjectService
from bq_meta.service.table_service import TableService
from bq_meta.service.version_service import VersionService
from bq_meta.util import table_utils
from bq_meta.window import Window


@click.command()
@click.argument("full_table_id", required=False)
@click.option("--raw", help="View raw response from the BigQuery for specific 'FULL_TABLE_ID'", is_flag=True)
@click.option("--init", help="Initialize 'bq-meta' configuration", is_flag=True)
@click.option("--info", help="Print info of currently used account", is_flag=True)
@click.option("--fetch-projects", help="Fetch available google projects", is_flag=True)
@click.version_option()
def cli(
    full_table_id: Optional[str],
    raw: bool,
    init: bool,
    info: bool,
    fetch_projects: bool,
):
    """BiqQuery metadata"""

    ctx = click.get_current_context()
    console = Console(theme=const.theme, soft_wrap=True, force_interactive=True)
    config = Config()
    bq_client = BqClient(console, config)
    project_service = ProjectService(console, config, bq_client)
    table_service = TableService(console, config, bq_client, project_service)
    history_service = HistoryService(console, config, table_service)
    window = Window(console, config, history_service, table_service)
    table = None
    
    if os.path.exists(const.BQ_META_HOME):
        version_service = VersionService()
        version_service.update_config(config)

    if init:
        initialize(config, console, project_service)
        ctx.exit()
    elif not os.path.exists(const.BQ_META_HOME):
        panel = Panel(
            title="Not initialized, run",
            renderable=Text("bq-meta --init"),
            expand=False,
            padding=(1, 3),
            border_style=const.request_style,
        )
        console.print(panel)
        ctx.exit()
    elif info:
        console.print(output.get_config_info(config))
        ctx.exit()
    elif fetch_projects:
        project_service.fetch_projects()
        ctx.exit()
    elif full_table_id:
        table_ref = TableReference.from_string(full_table_id.replace(":", "."))
        project_id = table_ref.project
        dataset_id = table_ref.dataset_id
        table_id = table_ref.table_id
        table = table_service.get_table(project_id, dataset_id, table_id)
        if raw:
            console.print_json(table_utils.get_properties(table))
            ctx.exit()

    window.live_window(table)



from enum import Enum
import webbrowser
from datetime import datetime
from typing import Optional
import click

import readchar
from bq_meta import const, output
from bq_meta.config import Config
from bq_meta.service.history_service import HistoryService
from bq_meta.service.table_service import TableService
from google.cloud import bigquery
from rich.console import Console
from rich.live import Live
from rich.panel import Panel
from rich.layout import Layout
from rich.console import RenderableType

from bq_meta.util.rich_utils import flash_panel


class View(Enum):
    empty = 1
    table = 2
    schema = 3


class Window:
    def __init__(
        self,
        console: Console,
        config: Config,
        history_service: HistoryService,
        table_service: TableService,
    ):
        self.console = console
        self.config = config
        self.history_service = history_service
        self.table_service = table_service
        self.table: Optional[bigquery.Table] = None
        self.project_id: Optional[str] = None
        self.dataset_id: Optional[str] = None
        self.layout: Layout = Layout()
        self.content: Optional[RenderableType] = None
        self.view: View = View.empty
        self.subtitle: Optional[str] = None

    def live_window(self, table: Optional[bigquery.Table]):
        if table:
            self._update_table(table)
            self._update_view(View.table)
        else:
            self._update_view(View.empty)

        with Live(self.layout, auto_refresh=False, screen=True, transient=True) as live:
            self._loop(live)

    def _update_panel(self, live: Live) -> None:
        content_layout = Layout()
        header = output.header_layout(self.config)
        if self.content:
            content = Layout(self.content, name="content")
        else:
            content = Layout(name="content", visible=False)
        content_layout.split_column(header, content)
        now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        panel = Panel(
            title=now,
            title_align="right",
            subtitle=self.subtitle,
            renderable=content_layout,
            border_style=const.border_style,
        )
        self.layout = Layout(panel)
        self.panel = panel
        live.update(self.layout, refresh=True)

    def _update_table(self, table: bigquery.Table):
        """
        Update table identifiers to the state, used when listing datasets or tables
        Save table to the history
        """
        if table:
            self.table = table
            self.project_id = table.project
            self.dataset_id = table.dataset_id
            self.history_service.save_table(table)

    def _update_view(self, view: View):
        """
        Set view, subtitle and content
        """
        if view == View.empty:
            self.view = view
            self.subtitle = "open (o) | history (h) | quit (q)"
        elif view == View.table and self.table:
            self.view = view
            self.subtitle = "open (o) | refresh (r) | schema (s) | console (c) | history (h) | quit (q)"
            self.content = output.get_table_output(self.table)
        elif view == View.schema and self.table:
            self.view = view
            self.subtitle = "open (o) | refresh (r) | table (t) | console (c) | history (h) | quit (q)"
            self.content = output.get_schema_output(self.table)

    def _loop(self, live: Live):
        """
        Loop listening on specific keypress, updating live CLI
        """

        self._update_panel(live)
        char = readchar.readkey()

        # List projects / Open new table
        if char == "1" or char == "o":
            table = self.table_service.get_table(live=live)
            if table:
                self._update_table(table)
                self._update_view(View.table)

        # List datasets
        elif char == "2":
            table = self.table_service.get_table(self.project_id, live=live)
            if table:
                self._update_table(table)
                self._update_view(View.table)

        # List tables
        elif char == "3":
            table = self.table_service.get_table(self.project_id, self.dataset_id, live=live)
            if table:
                self._update_table(table)
                self._update_view(View.table)

        # List history
        elif char == "h":
            table = self.history_service.pick_table(live)
            if table:
                self._update_table(table)
                self._update_view(View.table)

        # Refresh view
        elif char == "r" and self.table:
            table = self.table_service.get_fresh_table(self.table)
            flash_panel(live, self.layout, self.panel)
            if table:
                self.table = table
                self._update_view(self.view)

        # Show schema
        elif char == "s":
            if self.table:
                self._update_view(View.schema)

        # Show table
        elif char == "t":
            if self.table:
                self._update_view(View.table)

        # Open table in the google console
        elif char == "c":
            if self.table:
                url = f"https://console.cloud.google.com/bigquery?p={self.table.project}&d={self.table.dataset_id}&t={self.table.table_id}&page=table"
                webbrowser.open(url)

        # Quit program
        elif char == "q":
            live.stop()
            click.get_current_context().exit()

        # Infinite loop, until quitted (q)
        self._loop(live)



from pathlib import Path

from rich.console import Console, Group, NewLine
from rich.panel import Panel
from rich.prompt import Prompt
from rich.text import Text

from bq_meta import const
from bq_meta.auth import Auth
from bq_meta.config import Config
from bq_meta.service.project_service import ProjectService


def initialize(config: Config, console: Console, project_service: ProjectService):
    bq_meta_home = Prompt.ask(
        Text("", style=const.darker_style).append("Enter bq_meta home path", style=const.request_style),
        default=const.DEFAULT_BQ_META_HOME,
        console=console,
    )

    Path(bq_meta_home).mkdir(parents=True, exist_ok=True)
    _print_created(console, bq_meta_home)
    Path(const.BQ_META_CONFIG).touch()
    config.write_default()
    _print_created(console, const.BQ_META_CONFIG)
    Path(const.BQ_META_PROJECTS).touch()
    _print_created(console, const.BQ_META_PROJECTS)
    Path(const.BQ_META_HISTORY).touch()
    _print_created(console, const.BQ_META_HISTORY)

    Prompt.ask(
        Text("", style=const.darker_style).append("Login to google account", style=const.request_style),
        choices=None,
        default="Press enter",
        console=console,
    )
    auth = Auth(config, console)
    auth.login()

    project_service.fetch_projects()

    if bq_meta_home != const.DEFAULT_BQ_META_HOME:
        group = Group(Text(f"export BQ_META_HOME={bq_meta_home}"))
        console.print(
            NewLine(),
            Panel(
                title="Export following to your environment",
                renderable=group,
                expand=False,
                padding=(1, 3),
                border_style=const.request_style,
            ),
        )


def _print_created(console: Console, path: str):
    console.print(Text("Created", style=const.info_style).append(f": {path}", style=const.darker_style))



import json
from typing import List

from google.cloud import bigquery
from google.cloud.bigquery.schema import SchemaField
from rich.table import Table
from rich.text import Text
from rich.tree import Tree


def scheme_table(schema: List[SchemaField], table: Table):
    for field in schema:
        table.add_row(field.field_type, field.mode)
        if field.field_type == "RECORD":
            scheme_table(field.fields, table)
    return


def scheme_tree(schema: List[SchemaField], tree: Tree):
    for field in schema:
        node = tree.add(Text(f"{field.name}"))
        if field.field_type == "RECORD":
            scheme_tree(field.fields, node)
    return


def get_properties(table: bigquery.Table) -> str:
    return json.dumps(table._properties, indent=2)



from typing import List, Optional

from bq_meta import const
from bq_meta.config import Config
from bq_meta.service.table_service import TableService
from bq_meta.util import bash_util
from google.cloud.bigquery.table import Table, TableReference
from rich.console import Console
from rich.live import Live


class HistoryService:
    def __init__(self, console: Console, config: Config, table_service: TableService) -> None:
        self.console = console
        self.config = config
        self.table_service = table_service
        self.history_path = const.BQ_META_HISTORY

    def list_history(self) -> List[str]:
        projects = open(self.history_path, "r").read().splitlines()
        return projects

    def save_table(self, table: Table):
        history = self.list_history()
        try:
            history.remove(table.full_table_id)
        except ValueError:
            pass
        history.append(table.full_table_id)
        with open(self.history_path, "w") as f:
            f.write("\n".join(history))

    def pick_table(self, live: Optional[Live]) -> Optional[Table]:
        history = self.list_history()
        from_history = bash_util.pick_one(history, live)
        table = None
        if from_history:
            table_ref = TableReference.from_string(from_history.replace(":", "."))
            project_id = table_ref.project
            dataset_id = table_ref.dataset_id
            table_id = table_ref.table_id
            table = self.table_service.get_table(project_id, dataset_id, table_id, live)
        return table



from pathlib import Path
import os
import yaml
from rich.style import Style
from rich.theme import Theme
from rich.box import Box

DEFAULT_BQ_META_HOME = f"{Path.home()}/.config/bq-meta"

BQ_META_HOME = os.getenv("BQ_META_HOME", DEFAULT_BQ_META_HOME)
BQ_META_CONFIG = f"{BQ_META_HOME}/config.yaml"
BQ_META_PROJECTS = f"{BQ_META_HOME}/projects"
BQ_META_HISTORY = f"{BQ_META_HOME}/history"

BQ_META_DISABLE_COLORS = os.getenv("BQ_META_DISABLE_COLORS", "False").lower() in ("true", "1", "t")
BQ_META_SKIN = os.getenv("BQ_META_SKIN")

default_skin = {
    "request": "gold3",
    "info": "green",
    "error": "red",
    "border": "grey50",
    "time": "yellow",
    "darker": "grey27",
    "alternate": "grey50",
    "link": "light_sky_blue1",
    "key": "steel_blue1",
}

skin = default_skin

if BQ_META_SKIN and os.path.isfile(BQ_META_SKIN):
    BQ_META_skin = yaml.safe_load(open(BQ_META_SKIN, "r"))
    skin = {**skin, **BQ_META_skin}

request_style = Style(color=skin["request"])
info_style = Style(color=skin["info"])
error_style = Style(color=skin["error"])
border_style = Style(color=skin["border"])
time_style = Style(color=skin["time"])
darker_style = Style(color=skin["darker"])
alternate_style = Style(color=skin["alternate"])
link_style = Style(color=skin["link"])
key_style = Style(color=skin["key"])

theme = Theme(
    {
        "progress.elapsed": darker_style,
        "prompt.default": darker_style,
        "prompt.choices": "default",
        "rule.line": darker_style,
        "repr.path": darker_style,
        "repr.filename": darker_style,
        "json.brace": "none",
        "status.spinner": "none",
        "progress.spinner": "none",
        "repr.number": "none",
        "tree.line": darker_style,
    }
)

equal_box: Box = Box(
    """\
  = 
  = 
  = 
  = 
  = 
  = 
  = 
  = 
"""
)

FZF_SEPARATOR = " ~ "
